// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1997, 2001.
// ---------------------------------------------------------------------------
//	$Id: winsound.cpp,v 1.26 2002/05/31 09:45:21 cisc Exp $

#include "headers.h"
#include "WinSound.h"
	#include "misc.h"
#include "pc88/config.h"
#include "status.h"
#include "soundds.h"
#include "soundds2.h"
#include "soundwo.h"
#include "soundmon.h"

//#define LOGNAME "winsound"
#include "diag.h"

using namespace PC8801;
using namespace WinSoundDriver;

// ---------------------------------------------------------------------------
//	\z/
//
WinSound::WinSound()
: driver(0), hmmio(0)
{
	soundmon = 0;
	dumpstate = idle;
	useds2 = true;
}

WinSound::~WinSound()
{
	DumpEnd();
	Cleanup();
}

// ---------------------------------------------------------------------------
//	
//
bool WinSound::Init
(PC88* pc, HWND hwindow, uint rate, uint buflen)
{
	currentrate = 100;
	currentbuflen = 0;
	hwnd = hwindow;
	
	if (!Sound::Init(pc, 8000, 0))
		return false;
	return true;
}

// ---------------------------------------------------------------------------
//	㏈
//
void WinSound::Cleanup()
{
	if (driver)
	{
		driver->Cleanup();
		delete driver;
		driver = 0;
	}
	Sound::Cleanup();
}

// ---------------------------------------------------------------------------
//	EĐ[gύX
//
bool WinSound::ChangeRate(uint rate, uint buflen, bool waveout)
{
	if (currentrate != rate || currentbuflen != buflen || wodrv != waveout)
	{
		if (dumpstate != idle)
		{
			statusdisplay.Show(70, 3000, "wav ỏݒ̕ύX͂ł܂");
			return false;
		}

		samprate = rate;
		currentrate = rate;
		currentbuflen = buflen;
		wodrv = waveout;
		
		if (rate < 8000)
		{
			rate = 100; samprate = 0;
		}
		
		// DirectSound: TvO[g * obt@ / 2
		// waveOut:     TvO[g * obt@ * 2
		int bufsize;
		if (wodrv)
			bufsize = (samprate * buflen / 1000 * 1) & ~15;
		else
			bufsize = (samprate * buflen / 1000 / 2) & ~15;
		
		if (driver)
		{
			driver->Cleanup();
			delete driver;
			driver = 0;
		}

		if (rate < 1000)
			bufsize = 0;
			
		if (!SetRate(rate, bufsize))
			return false;
		
		if (bufsize > 0)
		{
			for (int i=0; i<1; i++)
			{
				if (wodrv)
				{
					driver = new DriverWO;
				}
				else if (useds2)
				{
					driver = new DriverDS2;
					statusdisplay.Show(200, 8000, "sounddrv: using Notify driven driver");
				}
				else
				{
					driver = new DriverDS;
					statusdisplay.Show(200, 8000, "sounddrv: using timer driven driver");
				}
				
				if (!driver || !driver->Init(GetSoundSource(), hwnd, samprate, 2, buflen))
				{
					delete driver; driver = 0;
					if (!wodrv && useds2)
					{
						useds2 = false;
						statusdisplay.Show(100, 3000, "IDirectSoundNotify ͎gpłȂ悤ł");
						i = -1;
					}
				}
			}
			if (!driver)
			{
				SetRate(rate, 0);
				statusdisplay.Show(70, 3000, "I[fBIfoCXgpł܂");
			}
		}
	}
	return true;
}

// ---------------------------------------------------------------------------
//	ݒXV
//
void WinSound::ApplyConfig(const Config* config)
{
	useds2 = !!(config->flag2 & Config::usedsnotify);


	bool wo = (config->flag2 & Config::usewaveoutdrv) != 0;
	ChangeRate(config->sound, config->soundbuffer, wo);
	
	if (driver)
		driver->MixAlways(0 != (config->flags & Config::mixsoundalways));
	
	Sound::ApplyConfig(config);
}

// ---------------------------------------------------------------------------
//	_vJn
//
bool WinSound::DumpBegin(char* filename)
{
	if (hmmio)
		return false;
	if (!driver)
	{
		statusdisplay.Show(70, 3000, "[hł PCM ̏o͍s܂");
		return false;
	}

	memset(&ckparent, 0, sizeof(MMCKINFO));
	memset(&ckdata, 0, sizeof(MMCKINFO));

	// mmioOpen
	hmmio = mmioOpen(filename, NULL, MMIO_CREATE | MMIO_WRITE | MMIO_ALLOCBUF); 
	if (!hmmio)
		return false;
	
	// WAVE chunk
	ckparent.fccType = mmioFOURCC('W', 'A', 'V', 'E');
	if (mmioCreateChunk(hmmio, &ckparent, MMIO_CREATERIFF))
	{
		mmioClose(hmmio, 0);
		hmmio = 0;
		return false;
	}

	// fmt chunk
	MMCKINFO cksub;
	memset(&cksub, 0, sizeof(MMCKINFO));
	cksub.ckid = mmioFOURCC('f', 'm', 't', ' ');
	mmioCreateChunk(hmmio, &cksub, 0);
	
	WAVEFORMATEX format;
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.nChannels = 2;
	format.nSamplesPerSec = samprate;
	format.wBitsPerSample = 16;
	format.nAvgBytesPerSec = format.nChannels * format.nSamplesPerSec * 2;
	format.nBlockAlign = format.nChannels * 2;
	format.cbSize = 0;

	mmioWrite(hmmio, HPSTR(&format), sizeof(format));
	mmioAscend(hmmio, &cksub, 0);

	// data chunk
	ckdata.ckid = mmioFOURCC('d', 'a', 't', 'a');
	mmioCreateChunk(hmmio, &ckdata, 0);
	
	FillWhenEmpty(false);
	dumpstate = standby;
	dumpedsample = 0;
	statusdisplay.Show(100, 0, "^ҋ@`");
	return true;
}

// ---------------------------------------------------------------------------
//	_vI
//
bool WinSound::DumpEnd()
{
	if (dumpstate != idle)
	{
		dumpstate = idle;
		CriticalSection::Lock lock(cs);

		if (ckdata.dwFlags & MMIO_DIRTY) 
			mmioAscend(hmmio, &ckdata, 0);
		if (ckparent.dwFlags & MMIO_DIRTY)
			mmioAscend(hmmio, &ckparent, 0);
		if (hmmio)
			mmioClose(hmmio, 0), hmmio = 0;
		FillWhenEmpty(true);

		int curtime = dumpedsample / currentrate;
		statusdisplay.Show(100, 2500, "^` [%.2d:%.2d]", curtime/60, curtime%60);
	}
	return true;
}

// ---------------------------------------------------------------------------
//	̏ꍇ
//
int WinSound::Get(Sample* dest, int samples)
{
	samples = Sound::Get(dest, samples);
	if (soundmon)
		soundmon->Mix(dest, samples);
	if (dumpstate != idle)
	{
		CriticalSection::Lock lock(cs);
		if (dumpstate != idle)
		{
			Dump(dest, samples);
		}
	}
	return samples;
}

void WinSound::Dump(Sample* dest, int samples)
{
	if (dumpstate == standby)
	{
		int i;
		uint32* s = (uint32*) dest;
		for (i=0; i<samples && !*s; i++, s++)
			;
		dest += i*2, samples -= i;
		if (samples > 0)
			dumpstate = dumping;
	}

	if (samples)
	{
		mmioWrite(hmmio, (char*) dest, samples * sizeof(Sample) * 2);
		int prevtime = dumpedsample / currentrate;
		dumpedsample += samples;
		int curtime = dumpedsample / currentrate;
		if (prevtime != curtime)
		{
			statusdisplay.Show(101, 0, "^` [%.2d:%.2d]", curtime/60, curtime%60);
		}
	}
}
